from reseasdk.version import __version__
from reseasdk.main import main

import reseasdk.helpers
import reseasdk.validators

command_list = [
    'build',
    'test',
    'clean',
    'log',
    'new',
    'version'
]
